@@ -1,6 +1,6 @@
 #!/usr/bin/python
 # -*- coding: utf-8 -*-
-from abra.config import DEFAULT_ALPHA
+from abra.config import DEFAULT_ALPHA, logger
 from abra.mixin import InitRepr
 from statsmodels.stats.api import DescrStatsW, CompareMeans
 from statsmodels.distributions.empirical_distribution import ECDF
@@ -8,7 +8,6 @@ from statsmodels.stats.power import tt_ind_solve_power, zt_ind_solve_power
 from statsmodels.stats.proportion import proportions_ztest, binom_test
 from scipy.stats import norm
 from scipy import optimize
-import logging
 from pandas import DataFrame
 import numpy as np
 
@@ -162,6 +161,7 @@ def estimate_experiment_sample_sizes(
             sample_size_ratio=.5
         )
     )
+    # [8590, 4295]
     """
     if statistic in ('t', 'z'):
         # std_control and/or std_variation are in *args, or **kwargs
@@ -470,7 +470,7 @@ Summary:
 𝝈² : {:1.4f}""".format(header, self.nobs, self.mean, self.var)
 
     def permute(self):
-        return np.random.choice(self.data, self.nobs)
+        return np.random.choice(self.data, int(self.nobs))
 
     def sort(self):
         if not hasattr(self, '_sorted'):
@@ -537,7 +537,7 @@ Summary:
             _hdi = highest_density_interval(self.data, credible_mass)
             return (round(_hdi[0], 4), round(_hdi[1], 4))
         except Exception as e:
-            logging.warn(e)
+            logger.warn(e)
             return (None, None)
 
     def hist(self, ref_val=None, *hist_args, **hist_kwargs):
@@ -547,7 +547,7 @@ Summary:
 
         """
         from matplotlib import pyplot as plt
-        pl = plt.hist(self.data, *hist_args, **hist_kwargs)
+        pl = plt.hist(self.data.astype(float), *hist_args, **hist_kwargs)
         if ref_val is not None:
             plt.axvline(ref_val, c='gray', linestyle='--', linewidth=2)
         return pl
@@ -634,7 +634,7 @@ class MeanComparison(CompareMeans):
 
 class ProportionComparison(MeanComparison):
     """
-    Class for comparing the propritions of two sample distributions, provides a number
+    Class for comparing the proportions of two sample distributions, provides a number
     of helpful summary statistics about the comparison. In order to use the
     z-distribution, we assume normality or proportions and thus, by proxy, adequate
     sample sizes (i.e. > 30).
@@ -661,7 +661,7 @@ class ProportionComparison(MeanComparison):
         # to use Normal approx, must have large N
         if nobs < 30:
             warning = 'Normality assumption violated, at least 30 observations required. Smallest sample size is {}'.format(nobs)
-            logging.warn(warning)
+            logger.warn(warning)
             self.warnings.append(warning)
 
         self.variance_assumption = variance_assumption
@@ -684,9 +684,11 @@ class ProportionComparison(MeanComparison):
         s_1 = sum(self.d1.data)
         n_2 = self.d2.nobs
         s_2 = sum(self.d2.data)
-        return proportions_ztest([s_1, s_2], [n_1, n_2],
-                                 alternative=self.test_direction,
-                                 prop_var=prop_var)
+        return proportions_ztest(
+            [s_1, s_2], [n_1, n_2],
+            alternative=self.test_direction,
+            prop_var=prop_var
+        )
 
 
 class RateComparison(MeanComparison):
@@ -743,6 +745,13 @@ class RateComparison(MeanComparison):
         """
         Run the rates comparison hyptothesis test. Uses the W5 statistic defined
         in Gu et al., 2008
+
+        Returns
+        -------
+        W : float
+            The W5 statistic from Gu et al., 2008
+        p_value : float
+            The p-value associated with W
         """
         X1, X2 = self.d2.sum, self.d1.sum
         t1, t2 = self.d2.nobs, self.d1.nobs
@@ -805,7 +814,6 @@ def highest_density_interval(samples, mass=.95):
     -------
     hdi: tuple(float)
         The lower and upper bounds of the highest density interval
-
     """
     _samples = np.asarray(sorted(samples))
     n = len(_samples)
@@ -823,128 +831,157 @@ def highest_density_interval(samples, mass=.95):
     return hdi_min, hdi_max
 
 
-class Traces(object):
+class BootstrapStatisticComparison(MeanComparison):
     """
-    Container class for analyzing the results of Bayesian inference procedure.
+    Class for comparing a bootstrapped test statistic for two samples. Provides
+    a number of helpful summary statistics about the comparison.
 
     Parameters
     ----------
-    traces: dict
-        Key-value pairs of parameters:samples, extracted from a Bayesian inference
-        procedure.
-    burnin: int
-        We ignore the first `burnin` samples to avoid any autocorrelation. This
-        is particularly helpful when samples are produced by MCMC.
+    samples_a : Samples instance
+        Group a samples
+    samples_b : Samples instance
+        Group b samples
+    alpha : float in (0, 1)
+        The assumed Type I error
+    hypothesis : str
+        Defines the assumed alternative hypothesis. Can be :
+            'larger'
+            'smaller'
+            'unequal' (i.e. two-tailed test)
+    n_bootstraps : int
+        The number of bootstrap samples to draw use for estimates.
+    statistic_function : function
+        Function that returns a scalar test statistic when provided a sequence
+        of samples.
+
+    References
+    ----------
+    Efron, B. (1981). "Nonparametric estimates of standard error: The jackknife, the bootstrap and other methods". Biometrika. 68 (3): 589–599
     """
+    def __init__(self, n_bootstraps=1000, statistic_function=None, *args, **kwargs):
+        statistic_function = statistic_function if statistic_function else np.mean
+        statistic_name = statistic_function.__name__
+        super(BootstrapStatisticComparison, self).__init__(
+            test_statistic=f"{statistic_name}", *args, **kwargs
+        )
+        self.statistic_function = statistic_function
+        self.n_bootstraps = n_bootstraps
 
-    def __init__(self, traces):
-        self.variables = []
-        for k, v in list(traces.items()):
-            if k != "lp__":
-                self.variables.append(k)
-                setattr(self, k, Samples(v))
-        self.summarize()
 
-    def summarize(self):
-        prct = [2.5, 25, 50, 75, 97.5]
-        values = []
-        columns = []
-        for v in self.variables:
-            trace = getattr(self, v)
-            _mean = trace.mean
-            _hdi = trace.hdi()
-            _std = trace.std
-            _percentiles = trace.percentiles(prct)
-            values.append(np.r_[_mean, _hdi, _std, _percentiles])
-        columns = ['mean', 'hdi_lower', 'hdi_upper', 'std'] + ["{}%".format(p) for p in prct]
-        self._summary = DataFrame(values, columns=columns, index=self.variables)
+    def bootstrap_test(self):
+        """
+        Run the sample comparison hyptothesis test. Uses the bootstrapped sample statistics
+
+        Returns
+        -------
+        delta: float
+            The observed difference in test statistic
+        p_value : float
+            The p-value associated with delta
+        """
+        all_samples = np.concatenate([self.d1.data, self.d2.data]).astype(float)
+
+        d1_samples = np.random.choice(all_samples, (int(self.d1.nobs), self.n_bootstraps), replace=True)
+        d1_statistics = np.apply_along_axis(self.statistic_function, axis=0, arr=d1_samples)
+
+        d2_samples = np.random.choice(all_samples, (int(self.d2.nobs), self.n_bootstraps), replace=True)
+        d2_statistics = np.apply_along_axis(self.statistic_function, axis=0, arr=d2_samples)
+
+        control_bs_samples = np.random.choice(self.d2.data, (int(self.d2.nobs), self.n_bootstraps), replace=True)
+        control_statistics = np.apply_along_axis(self.statistic_function, axis=0, arr=control_bs_samples)
+        self.control_bootstrap = Samples(control_statistics, name='control')
+
+        variation_bs_samples = np.random.choice(self.d1.data, (int(self.d1.nobs), self.n_bootstraps), replace=True)
+        variation_statistics = np.apply_along_axis(self.statistic_function, axis=0, arr=variation_bs_samples)
+        self.variation_bootstrap = Samples(variation_statistics, name='variation')
+
+        # The null sampling distribution of test_statistic deltas
+        self.null_dist = Samples(d2_statistics - d1_statistics, name=f'{self.test_statistic}-null')
+
+        if self.hypothesis == 'larger':
+            p_val = 1 - self.null_dist.cdf(self.delta)
+        elif self.hypothesis == 'smaller':
+            p_val = self.null_dist.cdf(self.delta)
+        elif self.hypothesis == 'unequal':
+            p_val = 1 - self.null_dist.cdf(abs(self.delta))
+
+        return self.delta, p_val
+
+
+    def confidence_interval(self, alpha=.05):
+        """
+        Calculate the (1-alpha)-th confidence interval around the statistic delta.
+        Uses bootstrapped approximation the statistic sampling distribution.
+
+        Returns
+        -------
+        ci : tuple (lo, hi)
+            the (1-alpha) % confidence interval around the statistic estimate.
+        """
+        return  self.deltas_dist.percentiles([100 * alpha, 100 * (1-alpha)])
 
     @property
-    def summary(self):
-        return self._summary
+    def deltas_dist(self):
+        if not hasattr(self, '_deltas_dist'):
+            d1_samples = np.random.choice(self.d1.data, (int(self.d1.nobs), self.n_bootstraps), replace=True)
+            d1_statistics = np.apply_along_axis(self.statistic_function, axis=0, arr=d1_samples)
 
-    def plot(self, variable, label=None,
-             color=None, ref_val=None, alpha=.25,
-             bins=None, title=None,
-             hdi=None, outfile=None,
-             ref_color=None):
+            d2_samples = np.random.choice(self.d2.data, (int(self.d2.nobs), self.n_bootstraps), replace=True)
+            d2_statistics = np.apply_along_axis(self.statistic_function, axis=0, arr=d2_samples)
+
+            self._deltas_dist = Samples(d1_statistics - d2_statistics, name=f'{self.test_statistic}-deltas')
+
+        return self._deltas_dist
+
+    @property
+    def delta(self):
         """
-        Plot the histogram of a variable trace
-
-        Parameters
-        ----------
-        variable : str
-            The name of one of self.variables to plot
-        label : str
-            Alternative label for the legend
-        ref_val : float
-            A reference value location at which to draw a vertical line
-        alpha : float in [0 1)
-            The transparency of the histogram
-        bins : int
-            The number of histogram bins
-        title : str
-            The title of the plot
-        hdi : float in [0, 1]
-            The amount of probability mass within the Highest Density Interval
-            to display on the histogram.
-        outfile : str
-            The name of an output file to save the figure to.
+        Delta is difference in test statistics
         """
-        from matplotlib import pyplot as plt  # lazy import
-        from abra.vis import plot_interval
+        return self.deltas_dist.mean
 
-        if variable not in self.variables:
-            print(self.variables)
-            raise ValueError('Variable `{}` not available'.format(variable))
+    @property
+    def delta_relative(self):
+        return self.delta / np.abs(self.statistic_function(self.d2.data))
 
-        label = label if label else variable
-        trace = getattr(self, variable)
-
-        if bins is None:
-            bins = int(len(trace.data) / 50.)
-
-        trace.hist(color=color, alpha=alpha, bins=bins, ref_val=ref_val, label=label)
-
-        if hdi is not None:  # highest density interval
-            median = round(trace.percentiles(50), 3)
-            _hdi = [round(h, 3) for h in trace.hdi(1 - hdi)]
-            plot_interval(*_hdi, middle=median, display_text=True, color=color, offset=5)
-
-        if title is None:
-            if ref_val is not None:
-                gt = round(100 * trace.prob_greater_than(ref_val))
-                title = " {}% < {} = {} < {}%".format(100 - gt, variable, ref_val, gt)
-            else:
-                title = ''
-        plt.title(title, fontsize=16)
-
-        if outfile:
-            plt.savefig(outfile)
+    @property
+    def power(self):
+        """
+        Return the statistical power of the current test. Uses
+        """
+        critical_value = self.null_dist.percentiles(100 * (1 - self.alpha))
+        return self.deltas_dist.prob_greater_than(critical_value)
 
 
-def extract_traces(inference_results, backend='pystan'):
+def highest_density_interval(samples, mass=.95):
     """
+    Determine the bounds of the interval of width `mass` with the highest density
+    under the distribution of samples.
+
     Parameters
-    ---------
-    inference_results: the output of a StanModel.sampling or .vb method (dict)
-        The results of a PyStan Bayesian inference procedure, holds samples
-        for target parameters
+    ----------
+    samples: list
+        The samples to compute the interval over
+    mass: float (0, 1)
+        The credible mass under the empricial distribution
 
     Returns
     -------
-    traces: Traces instance
-        a helper class for analyzing Bayesian inference results
+    hdi: tuple(float)
+        The lower and upper bounds of the highest density interval
     """
-    if backend == 'pystan':
-        if issubclass(inference_results.__class__, dict):
-            traces = {}
-            for ii in range(len(inference_results['sampler_param_names'])):
-                param_name = inference_results['sampler_param_names'][ii]
-                param_value = np.array(inference_results['sampler_params'][ii])
-                traces[param_name] = param_value
-        else:
-            traces = inference_results.extract()
-    else:
-        raise ValueError("Only 'pystan' backend currently supported.")
-    return Traces(traces)
+    _samples = np.asarray(sorted(samples))
+    n = len(_samples)
+
+    interval_idx_inc = int(np.floor(mass * n))
+    n_intervals = n - interval_idx_inc
+    interval_width = _samples[interval_idx_inc:] - _samples[:n_intervals]
+
+    if len(interval_width) == 0:
+        raise ValueError('Too few elements for interval calculation')
+
+    min_idx = np.argmin(interval_width)
+    hdi_min = _samples[min_idx]
+    hdi_max = _samples[min_idx + interval_idx_inc]
+    return hdi_min, hdi_max
